# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for version_info."""

import os
import re
from io import BytesIO, StringIO

import yaml

from .. import registry, tests, version_info_formats
from ..bzr.rio import read_stanzas
from ..version_info_formats.format_custom import (
    CustomVersionInfoBuilder,
    MissingTemplateVariable,
    NoTemplate,
)
from ..version_info_formats.format_python import PythonVersionInfoBuilder
from ..version_info_formats.format_rio import RioVersionInfoBuilder
from ..version_info_formats.format_yaml import YamlVersionInfoBuilder
from . import TestCaseWithTransport


class VersionInfoTestCase(TestCaseWithTransport):
    def create_branch(self):
        wt = self.make_branch_and_tree("branch")

        self.build_tree(["branch/a"])
        wt.add("a")
        wt.commit("a", rev_id=b"r1")

        self.build_tree(["branch/b"])
        wt.add("b")
        wt.commit("b", rev_id=b"r2")

        self.build_tree_contents([("branch/a", b"new contents\n")])
        wt.commit("\xe52", rev_id=b"r3")

        return wt

    def create_tree_with_dotted_revno(self):
        wt = self.make_branch_and_tree("branch")
        self.build_tree(["branch/a"])
        wt.add("a")
        wt.commit("a", rev_id=b"r1")

        other = wt.controldir.sprout("other").open_workingtree()
        self.build_tree(["other/b.a"])
        other.add(["b.a"])
        other.commit("b.a", rev_id=b"o2")

        os.chdir("branch")
        self.run_bzr("merge ../other")
        wt.commit("merge", rev_id=b"merge")

        wt.update(revision=b"o2")

        return wt


class TestVersionInfoRio(VersionInfoTestCase):
    def test_rio_null(self):
        wt = self.make_branch_and_tree("branch")

        bio = BytesIO()
        builder = RioVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(bio)
        val = bio.getvalue()
        self.assertContainsRe(val, b"build-date:")
        self.assertContainsRe(val, b"revno: 0")

    def test_rio_dotted_revno(self):
        wt = self.create_tree_with_dotted_revno()

        bio = BytesIO()
        builder = RioVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(bio)
        val = bio.getvalue()
        self.assertContainsRe(val, b"revno: 1.1.1")

    def regen_text(self, wt, **kwargs):
        bio = BytesIO()
        builder = RioVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        builder.generate(bio)
        val = bio.getvalue()
        return val

    def test_simple(self):
        wt = self.create_branch()

        val = self.regen_text(wt)
        self.assertContainsRe(val, b"build-date:")
        self.assertContainsRe(val, b"date:")
        self.assertContainsRe(val, b"revno: 3")
        self.assertContainsRe(val, b"revision-id: r3")

    def test_clean(self):
        wt = self.create_branch()
        val = self.regen_text(wt, check_for_clean=True)
        self.assertContainsRe(val, b"clean: True")

    def test_no_clean(self):
        wt = self.create_branch()
        self.build_tree(["branch/c"])
        val = self.regen_text(wt, check_for_clean=True)
        self.assertContainsRe(val, b"clean: False")

    def test_history(self):
        wt = self.create_branch()

        val = self.regen_text(wt, include_revision_history=True)
        self.assertContainsRe(val, b"id: r1")
        self.assertContainsRe(val, b"message: a")
        self.assertContainsRe(val, b"id: r2")
        self.assertContainsRe(val, b"message: b")
        self.assertContainsRe(val, b"id: r3")
        self.assertContainsRe(val, b"message: \xc3\xa52")

    def regen(self, wt, **kwargs):
        bio = BytesIO()
        builder = RioVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        builder.generate(bio)
        bio.seek(0)
        stanzas = list(read_stanzas(bio))
        self.assertEqual(1, len(stanzas))
        return stanzas[0]

    def test_rio_version_hook(self):
        def update_stanza(rev, stanza):
            stanza.add("bla", "bloe")

        RioVersionInfoBuilder.hooks.install_named_hook("revision", update_stanza, None)
        wt = self.create_branch()

        stanza = self.regen(wt)
        self.assertEqual(["bloe"], stanza.get_all("bla"))

    def get_one_stanza(self, stanza, key):
        new_stanzas = list(read_stanzas(BytesIO(stanza[key].encode("utf8"))))
        self.assertEqual(1, len(new_stanzas))
        return new_stanzas[0]

    def test_build_date(self):
        wt = self.create_branch()
        stanza = self.regen(wt)
        self.assertTrue("date" in stanza)
        self.assertTrue("build-date" in stanza)
        self.assertEqual(["3"], stanza.get_all("revno"))
        self.assertEqual(["r3"], stanza.get_all("revision-id"))

    def test_not_clean(self):
        wt = self.create_branch()
        self.build_tree(["branch/c"])
        stanza = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        self.assertEqual(["False"], stanza.get_all("clean"))

    def test_file_revisions(self):
        wt = self.create_branch()
        self.build_tree(["branch/c"])
        stanza = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        # This assumes it's being run against a tree that does not update the
        # root revision on every commit.
        file_rev_stanza = self.get_one_stanza(stanza, "file-revisions")
        self.assertEqual(["", "a", "b", "c"], file_rev_stanza.get_all("path"))
        self.assertEqual(
            ["r1", "r3", "r2", "unversioned"], file_rev_stanza.get_all("revision")
        )

    def test_revision_history(self):
        wt = self.create_branch()
        stanza = self.regen(wt, include_revision_history=True)
        revision_stanza = self.get_one_stanza(stanza, "revisions")
        self.assertEqual(["r1", "r2", "r3"], revision_stanza.get_all("id"))
        self.assertEqual(["a", "b", "\xe52"], revision_stanza.get_all("message"))
        self.assertEqual(3, len(revision_stanza.get_all("date")))

    def test_file_revisions_with_rename(self):
        # a was modified, so it should show up modified again
        wt = self.create_branch()
        self.build_tree(["branch/a", "branch/c"])
        wt.add("c")
        wt.rename_one("b", "d")
        stanza = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        file_rev_stanza = self.get_one_stanza(stanza, "file-revisions")
        self.assertEqual(["", "a", "b", "c", "d"], file_rev_stanza.get_all("path"))
        self.assertEqual(
            ["r1", "modified", "renamed to d", "new", "renamed from b"],
            file_rev_stanza.get_all("revision"),
        )

    def test_file_revisions_with_removal(self):
        wt = self.create_branch()
        self.build_tree(["branch/a", "branch/c"])
        wt.add("c")
        wt.rename_one("b", "d")

        wt.commit("modified", rev_id=b"r4")

        wt.remove(["c", "d"])
        os.remove("branch/d")
        stanza = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        file_rev_stanza = self.get_one_stanza(stanza, "file-revisions")
        self.assertEqual(["", "a", "c", "d"], file_rev_stanza.get_all("path"))
        self.assertEqual(
            ["r1", "r4", "unversioned", "removed"], file_rev_stanza.get_all("revision")
        )

    def test_revision(self):
        wt = self.create_branch()
        self.build_tree(["branch/a", "branch/c"])
        wt.add("c")
        wt.rename_one("b", "d")

        stanza = self.regen(
            wt,
            check_for_clean=True,
            include_file_revisions=True,
            revision_id=wt.last_revision(),
        )
        file_rev_stanza = self.get_one_stanza(stanza, "file-revisions")
        self.assertEqual(["", "a", "b"], file_rev_stanza.get_all("path"))
        self.assertEqual(["r1", "r3", "r2"], file_rev_stanza.get_all("revision"))


class TestVersionInfoYaml(VersionInfoTestCase):
    def test_yaml_null(self):
        wt = self.make_branch_and_tree("branch")

        bio = StringIO()
        builder = YamlVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(bio)
        val = bio.getvalue()
        self.assertContainsRe(val, "build-date:")
        self.assertContainsRe(val, "revno: '0'")

    def test_yaml_dotted_revno(self):
        wt = self.create_tree_with_dotted_revno()

        bio = StringIO()
        builder = YamlVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(bio)
        val = bio.getvalue()
        self.assertContainsRe(val, "revno: 1.1.1")

    def regen_text(self, wt, **kwargs):
        bio = StringIO()
        builder = YamlVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        builder.generate(bio)
        val = bio.getvalue()
        return val

    def test_simple(self):
        wt = self.create_branch()

        val = self.regen_text(wt)
        self.assertContainsRe(val, "build-date:")
        self.assertContainsRe(val, "date:")
        self.assertContainsRe(val, "revno: '3'")
        self.assertContainsRe(val, "revision-id: r3")

    def test_clean(self):
        wt = self.create_branch()
        val = self.regen_text(wt, check_for_clean=True)
        self.assertContainsRe(val, "clean: true")

    def test_no_clean(self):
        wt = self.create_branch()
        self.build_tree(["branch/c"])
        val = self.regen_text(wt, check_for_clean=True)
        self.assertContainsRe(val, "clean: false")

    def test_history(self):
        wt = self.create_branch()

        val = self.regen_text(wt, include_revision_history=True)
        self.assertContainsRe(val, "id: r1")
        self.assertContainsRe(val, "message: a")
        self.assertContainsRe(val, "id: r2")
        self.assertContainsRe(val, "message: ")
        self.assertContainsRe(val, "id: r3")
        self.assertContainsRe(val, re.escape('message: "\\xE52"'))

    def regen(self, wt, **kwargs):
        bio = StringIO()
        builder = YamlVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        builder.generate(bio)
        bio.seek(0)
        return yaml.safe_load(bio)

    def test_yaml_version_hook(self):
        def update_stanza(rev, stanza):
            stanza["bla"] = "bloe"

        YamlVersionInfoBuilder.hooks.install_named_hook("revision", update_stanza, None)
        wt = self.create_branch()

        stanza = self.regen(wt)
        self.assertEqual("bloe", stanza["bla"])

    def test_build_date(self):
        wt = self.create_branch()
        stanza = self.regen(wt)
        self.assertTrue("date" in stanza)
        self.assertTrue("build-date" in stanza)
        self.assertEqual("3", stanza["revno"])
        self.assertEqual("r3", stanza["revision-id"])

    def test_not_clean(self):
        wt = self.create_branch()
        self.build_tree(["branch/c"])
        stanza = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        self.assertEqual(False, stanza["clean"])

    def test_file_revisions(self):
        wt = self.create_branch()
        self.build_tree(["branch/c"])
        stanza = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        # This assumes it's being run against a tree that does not update the
        # root revision on every commit.
        file_rev_stanza = stanza["file-revisions"]
        self.assertEqual(["", "a", "b", "c"], [r["path"] for r in file_rev_stanza])
        self.assertEqual(
            ["r1", "r3", "r2", "unversioned"], [r["revision"] for r in file_rev_stanza]
        )

    def test_revision_history(self):
        wt = self.create_branch()
        stanza = self.regen(wt, include_revision_history=True)
        revision_stanza = stanza["revisions"]
        self.assertEqual(["r1", "r2", "r3"], [r["id"] for r in revision_stanza])
        self.assertEqual(["a", "b", "\xe52"], [r["message"] for r in revision_stanza])
        self.assertEqual(3, len([r["date"] for r in revision_stanza]))

    def test_file_revisions_with_rename(self):
        # a was modified, so it should show up modified again
        wt = self.create_branch()
        self.build_tree(["branch/a", "branch/c"])
        wt.add("c")
        wt.rename_one("b", "d")
        stanza = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        file_rev_stanza = stanza["file-revisions"]
        self.assertEqual(["", "a", "b", "c", "d"], [r["path"] for r in file_rev_stanza])
        self.assertEqual(
            ["r1", "modified", "renamed to d", "new", "renamed from b"],
            [r["revision"] for r in file_rev_stanza],
        )

    def test_file_revisions_with_removal(self):
        wt = self.create_branch()
        self.build_tree(["branch/a", "branch/c"])
        wt.add("c")
        wt.rename_one("b", "d")

        wt.commit("modified", rev_id=b"r4")

        wt.remove(["c", "d"])
        os.remove("branch/d")
        stanza = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        file_rev_stanza = stanza["file-revisions"]
        self.assertEqual(["", "a", "c", "d"], [r["path"] for r in file_rev_stanza])
        self.assertEqual(
            ["r1", "r4", "unversioned", "removed"],
            [r["revision"] for r in file_rev_stanza],
        )

    def test_revision(self):
        wt = self.create_branch()
        self.build_tree(["branch/a", "branch/c"])
        wt.add("c")
        wt.rename_one("b", "d")

        stanza = self.regen(
            wt,
            check_for_clean=True,
            include_file_revisions=True,
            revision_id=wt.last_revision(),
        )
        file_rev_stanza = stanza["file-revisions"]
        self.assertEqual(["", "a", "b"], [r["path"] for r in file_rev_stanza])
        self.assertEqual(["r1", "r3", "r2"], [r["revision"] for r in file_rev_stanza])

    def test_no_wt(self):
        wt = self.create_branch()
        self.build_tree(["branch/a", "branch/c"])
        wt.add("c")
        wt.rename_one("b", "d")

        bio = StringIO()
        builder = YamlVersionInfoBuilder(
            wt.branch,
            working_tree=None,
            check_for_clean=True,
            include_file_revisions=True,
            revision_id=None,
        )
        builder.generate(bio)
        bio.seek(0)
        stanza = yaml.safe_load(bio)
        self.assertEqual([], stanza["file-revisions"])


class PythonVersionInfoTests(VersionInfoTestCase):
    def test_python_null(self):
        wt = self.make_branch_and_tree("branch")

        sio = StringIO()
        builder = PythonVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(sio)
        val = sio.getvalue()
        self.assertContainsRe(val, "'revision_id': None")
        self.assertContainsRe(val, "'revno': '0'")
        self.assertNotContainsString(val, "\n\n\n\n")

    def test_python_dotted_revno(self):
        wt = self.create_tree_with_dotted_revno()

        sio = StringIO()
        builder = PythonVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(sio)
        val = sio.getvalue()
        self.assertContainsRe(val, "'revno': '1.1.1'")

    def regen(self, wt, **kwargs):
        """Create a test module, import and return it."""
        builder = PythonVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        outf = StringIO()
        builder.generate(outf)
        local_vars = {}
        exec(outf.getvalue(), {}, local_vars)
        return local_vars

    def test_python_version(self):
        wt = self.create_branch()

        tvi = self.regen(wt)
        self.assertEqual("3", tvi["version_info"]["revno"])
        self.assertEqual(b"r3", tvi["version_info"]["revision_id"])
        self.assertTrue("date" in tvi["version_info"])
        self.assertEqual(None, tvi["version_info"]["clean"])

        tvi = self.regen(wt, check_for_clean=True)
        self.assertTrue(tvi["version_info"]["clean"])

        self.build_tree(["branch/c"])
        tvi = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        self.assertFalse(tvi["version_info"]["clean"])
        self.assertEqual(["", "a", "b", "c"], sorted(tvi["file_revisions"].keys()))
        self.assertEqual(b"r3", tvi["file_revisions"]["a"])
        self.assertEqual(b"r2", tvi["file_revisions"]["b"])
        self.assertEqual("unversioned", tvi["file_revisions"]["c"])
        os.remove("branch/c")

        tvi = self.regen(wt, include_revision_history=True)

        rev_info = [
            (rev, message) for rev, message, timestamp, timezone in tvi["revisions"]
        ]
        self.assertEqual([(b"r1", "a"), (b"r2", "b"), (b"r3", "\xe52")], rev_info)

        # a was modified, so it should show up modified again
        self.build_tree(["branch/a", "branch/c"])
        wt.add("c")
        wt.rename_one("b", "d")
        tvi = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        self.assertEqual(["", "a", "b", "c", "d"], sorted(tvi["file_revisions"].keys()))
        self.assertEqual("modified", tvi["file_revisions"]["a"])
        self.assertEqual("renamed to d", tvi["file_revisions"]["b"])
        self.assertEqual("new", tvi["file_revisions"]["c"])
        self.assertEqual("renamed from b", tvi["file_revisions"]["d"])

        wt.commit("modified", rev_id=b"r4")
        wt.remove(["c", "d"])
        os.remove("branch/d")
        tvi = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        self.assertEqual(["", "a", "c", "d"], sorted(tvi["file_revisions"].keys()))
        self.assertEqual(b"r4", tvi["file_revisions"]["a"])
        self.assertEqual("unversioned", tvi["file_revisions"]["c"])
        self.assertEqual("removed", tvi["file_revisions"]["d"])


class CustomVersionInfoTests(VersionInfoTestCase):
    def test_custom_null(self):
        sio = StringIO()
        wt = self.make_branch_and_tree("branch")
        builder = CustomVersionInfoBuilder(
            wt.branch, working_tree=wt, template="revno: {revno}"
        )
        builder.generate(sio)
        self.assertEqual("revno: 0", sio.getvalue())

        builder = CustomVersionInfoBuilder(
            wt.branch, working_tree=wt, template="{revno} revid: {revision_id}"
        )
        # revision_id is not available yet
        self.assertRaises(MissingTemplateVariable, builder.generate, sio)

    def test_custom_dotted_revno(self):
        sio = StringIO()
        wt = self.create_tree_with_dotted_revno()
        builder = CustomVersionInfoBuilder(
            wt.branch, working_tree=wt, template="{revno} revid: {revision_id}"
        )
        builder.generate(sio)
        self.assertEqual("1.1.1 revid: o2", sio.getvalue())

    def regen(self, wt, tpl, **kwargs):
        sio = StringIO()
        builder = CustomVersionInfoBuilder(
            wt.branch, working_tree=wt, template=tpl, **kwargs
        )
        builder.generate(sio)
        val = sio.getvalue()
        return val

    def test_build_date(self):
        wt = self.create_branch()

        val = self.regen(wt, 'build-date: "{build_date}"\ndate: "{date}"')
        self.assertContainsRe(val, 'build-date: "[0-9-+: ]+"')
        self.assertContainsRe(val, 'date: "[0-9-+: ]+"')

    def test_revno(self):
        wt = self.create_branch()
        val = self.regen(wt, "revno: {revno}")
        self.assertEqual(val, "revno: 3")

    def test_revision_id(self):
        wt = self.create_branch()
        val = self.regen(wt, "revision-id: {revision_id}")
        self.assertEqual(val, "revision-id: r3")

    def test_clean(self):
        wt = self.create_branch()
        val = self.regen(wt, "clean: {clean}", check_for_clean=True)
        self.assertEqual(val, "clean: 1")

    def test_not_clean(self):
        wt = self.create_branch()

        self.build_tree(["branch/c"])
        val = self.regen(wt, "clean: {clean}", check_for_clean=True)
        self.assertEqual(val, "clean: 0")
        os.remove("branch/c")

    def test_custom_without_template(self):
        builder = CustomVersionInfoBuilder(None)
        sio = StringIO()
        self.assertRaises(NoTemplate, builder.generate, sio)


class TestBuilder(version_info_formats.VersionInfoBuilder):
    pass


class TestVersionInfoFormatRegistry(tests.TestCase):
    def setUp(self):
        super().setUp()
        self.overrideAttr(version_info_formats, "format_registry", registry.Registry())

    def test_register_remove(self):
        registry = version_info_formats.format_registry
        registry.register("testbuilder", TestBuilder, "a simple test builder")
        self.assertIs(TestBuilder, registry.get("testbuilder"))
        self.assertEqual("a simple test builder", registry.get_help("testbuilder"))
        registry.remove("testbuilder")
        self.assertRaises(KeyError, registry.get, "testbuilder")



# Copyright (C) 2005-2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy.tests import TestCaseWithTransport


class TestRevisionInfo(TestCaseWithTransport):
    def check_output(self, output, *args):
        """Verify that the expected output matches what brz says.

        The output is supplied first, so that you can supply a variable
        number of arguments to bzr.
        """
        self.assertEqual(self.run_bzr(*args)[0], output)

    def test_revision_info(self):
        """Test that 'brz revision-info' reports the correct thing."""
        wt = self.make_branch_and_tree(".")

        # Make history with a non-mainline rev
        wt.commit("Commit one", rev_id=b"a@r-0-1")
        wt.commit("Commit two", rev_id=b"a@r-0-1.1.1")
        wt.set_parent_ids([b"a@r-0-1", b"a@r-0-1.1.1"])
        wt.branch.set_last_revision_info(1, b"a@r-0-1")
        wt.commit("Commit three", rev_id=b"a@r-0-2")

        # This is expected to work even if the working tree is removed
        wt.controldir.destroy_workingtree()

        # Expected return values
        values = {
            "1": "1 a@r-0-1\n",
            "1.1.1": "1.1.1 a@r-0-1.1.1\n",
            "2": "2 a@r-0-2\n",
        }

        # Make sure with no arg it defaults to the head
        self.check_output(values["2"], "revision-info")

        # Check the results of just specifying a numeric revision
        self.check_output(values["1"], "revision-info 1")
        self.check_output(values["1.1.1"], "revision-info 1.1.1")
        self.check_output(values["2"], "revision-info 2")
        self.check_output(values["1"] + values["2"], "revision-info 1 2")
        self.check_output(
            "    " + values["1"] + values["1.1.1"] + "    " + values["2"],
            "revision-info 1 1.1.1 2",
        )
        self.check_output(values["2"] + values["1"], "revision-info 2 1")

        # Check as above, only using the '--revision' syntax

        self.check_output(values["1"], "revision-info -r 1")
        self.check_output(values["1.1.1"], "revision-info --revision 1.1.1")
        self.check_output(values["2"], "revision-info -r 2")
        self.check_output(values["1"] + values["2"], "revision-info -r 1..2")
        self.check_output(
            "    " + values["1"] + values["1.1.1"] + "    " + values["2"],
            "revision-info -r 1..1.1.1..2",
        )
        self.check_output(values["2"] + values["1"], "revision-info -r 2..1")

        # Now try some more advanced revision specifications

        self.check_output(values["1"], "revision-info -r revid:a@r-0-1")
        self.check_output(values["1.1.1"], "revision-info --revision revid:a@r-0-1.1.1")

    def test_revision_info_explicit_branch_dir(self):
        """Test that 'brz revision-info' honors the '-d' option."""
        wt = self.make_branch_and_tree("branch")

        wt.commit("Commit one", rev_id=b"a@r-0-1")
        self.check_output("1 a@r-0-1\n", "revision-info -d branch")

    def test_revision_info_tree(self):
        # Make branch and checkout
        wt = self.make_branch_and_tree("branch")
        wt.commit("Commit one", rev_id=b"a@r-0-1")

        # Make checkout and move the branch forward
        wt.branch.create_checkout("checkout", lightweight=True)
        wt.commit("Commit two", rev_id=b"a@r-0-2")

        # Make sure the checkout gives the right answer for branch and
        # tree
        self.check_output("2 a@r-0-2\n", "revision-info -d checkout")
        self.check_output("1 a@r-0-1\n", "revision-info --tree -d checkout")

    def test_revision_info_tree_no_working_tree(self):
        # Make branch with no tree
        self.make_branch("branch")

        # Try getting the --tree revision-info
        out, err = self.run_bzr("revision-info --tree -d branch", retcode=3)
        self.assertEqual("", out)
        self.assertEqual('brz: ERROR: No WorkingTree exists for "branch".\n', err)

    def test_revision_info_not_in_history(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", None))], revision_id=b"A-id"
        )
        builder.build_snapshot([b"A-id"], [], revision_id=b"B-id")
        builder.build_snapshot([b"A-id"], [], revision_id=b"C-id")
        builder.finish_series()
        self.check_output(
            "  1 A-id\n??? B-id\n  2 C-id\n",
            "revision-info -d branch revid:A-id revid:B-id revid:C-id",
        )



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for eol conversion."""

import sys
from io import BytesIO

from ... import rules, status
from ...workingtree import WorkingTree
from .. import TestSkipped
from . import TestCaseWithWorkingTree

# Sample files
_sample_text = b"""hello\nworld\r\n"""
_sample_text_on_win = b"""hello\r\nworld\r\n"""
_sample_text_on_unix = b"""hello\nworld\n"""
_sample_binary = b"""hello\nworld\r\n\x00"""
_sample_clean_lf = _sample_text_on_unix
_sample_clean_crlf = _sample_text_on_win


# Lists of formats for each storage policy
_LF_IN_REPO = ["native", "lf", "crlf"]
_CRLF_IN_REPO = ["{}-with-crlf-in-repo".format(f) for f in _LF_IN_REPO]


class TestEolConversion(TestCaseWithWorkingTree):
    def setUp(self):
        # formats that don't support content filtering can skip these tests
        fmt = self.workingtree_format
        f = fmt.supports_content_filtering
        if f is None:
            raise TestSkipped(
                "format {} doesn't declare whether it "
                "supports content filtering, assuming not".format(fmt)
            )
        if not f():
            raise TestSkipped("format {} doesn't support content filtering".format(fmt))
        super().setUp()

    def patch_rules_searcher(self, eol):
        """Patch in a custom rules searcher with a given eol setting."""
        if eol is None:
            WorkingTree._get_rules_searcher = self.real_rules_searcher
        else:

            def custom_eol_rules_searcher(tree, default_searcher):
                return rules._IniBasedRulesSearcher(
                    [
                        "[name *]\n",
                        "eol={}\n".format(eol),
                    ]
                )

            WorkingTree._get_rules_searcher = custom_eol_rules_searcher

    def prepare_tree(self, content, eol=None):
        """Prepare a working tree and commit some content."""
        self.real_rules_searcher = self.overrideAttr(WorkingTree, "_get_rules_searcher")
        self.patch_rules_searcher(eol)
        t = self.make_branch_and_tree("tree1")
        self.build_tree_contents([("tree1/file1", content)])
        t.add("file1")
        t.commit("add file1")
        basis = t.basis_tree()
        basis.lock_read()
        self.addCleanup(basis.unlock)
        return t, basis

    def assertNewContentForSetting(
        self, wt, eol, expected_unix, expected_win, roundtrip
    ):
        """Clone a working tree and check the convenience content.

        If roundtrip is True, status and commit should see no changes.
        """
        if expected_win is None:
            expected_win = expected_unix
        self.patch_rules_searcher(eol)
        wt2 = wt.controldir.sprout("tree-{}".format(eol)).open_workingtree()
        # To see exactly what got written to disk, we need an unfiltered read
        with wt2.get_file("file1", filtered=False) as f:
            content = f.read()
        if sys.platform == "win32":
            self.assertEqual(expected_win, content)
        else:
            self.assertEqual(expected_unix, content)
        # Confirm that status thinks nothing has changed if the text roundtrips
        if roundtrip:
            status_io = BytesIO()
            status.show_tree_status(wt2, to_file=status_io)
            self.assertEqual(b"", status_io.getvalue())

    def assertContent(
        self, wt, basis, expected_raw, expected_unix, expected_win, roundtrip_to=None
    ):
        """Check the committed content and content in cloned trees.

        :param roundtrip_to: the set of formats (excluding exact) we
          can round-trip to or None for all
        """
        with basis.get_file("file1") as f:
            basis_content = f.read()
        self.assertEqual(expected_raw, basis_content)

        # No setting and exact should always roundtrip
        self.assertNewContentForSetting(
            wt, None, expected_raw, expected_raw, roundtrip=True
        )
        self.assertNewContentForSetting(
            wt, "exact", expected_raw, expected_raw, roundtrip=True
        )

        # Roundtripping is otherwise dependent on whether the original
        # text is clean - mixed line endings will prevent it. It also
        # depends on whether the format in the repository is being changed.
        if roundtrip_to is None:
            roundtrip_to = _LF_IN_REPO + _CRLF_IN_REPO
        self.assertNewContentForSetting(
            wt, "native", expected_unix, expected_win, "native" in roundtrip_to
        )
        self.assertNewContentForSetting(
            wt, "lf", expected_unix, expected_unix, "lf" in roundtrip_to
        )
        self.assertNewContentForSetting(
            wt, "crlf", expected_win, expected_win, "crlf" in roundtrip_to
        )
        self.assertNewContentForSetting(
            wt,
            "native-with-crlf-in-repo",
            expected_unix,
            expected_win,
            "native-with-crlf-in-repo" in roundtrip_to,
        )
        self.assertNewContentForSetting(
            wt,
            "lf-with-crlf-in-repo",
            expected_unix,
            expected_unix,
            "lf-with-crlf-in-repo" in roundtrip_to,
        )
        self.assertNewContentForSetting(
            wt,
            "crlf-with-crlf-in-repo",
            expected_win,
            expected_win,
            "crlf-with-crlf-in-repo" in roundtrip_to,
        )

    # Test binary files. These always roundtrip.

    def test_eol_no_rules_binary(self):
        wt, basis = self.prepare_tree(_sample_binary)
        self.assertContent(wt, basis, _sample_binary, _sample_binary, _sample_binary)

    def test_eol_exact_binary(self):
        wt, basis = self.prepare_tree(_sample_binary, eol="exact")
        self.assertContent(wt, basis, _sample_binary, _sample_binary, _sample_binary)

    def test_eol_native_binary(self):
        wt, basis = self.prepare_tree(_sample_binary, eol="native")
        self.assertContent(wt, basis, _sample_binary, _sample_binary, _sample_binary)

    def test_eol_lf_binary(self):
        wt, basis = self.prepare_tree(_sample_binary, eol="lf")
        self.assertContent(wt, basis, _sample_binary, _sample_binary, _sample_binary)

    def test_eol_crlf_binary(self):
        wt, basis = self.prepare_tree(_sample_binary, eol="crlf")
        self.assertContent(wt, basis, _sample_binary, _sample_binary, _sample_binary)

    def test_eol_native_with_crlf_in_repo_binary(self):
        wt, basis = self.prepare_tree(_sample_binary, eol="native-with-crlf-in-repo")
        self.assertContent(wt, basis, _sample_binary, _sample_binary, _sample_binary)

    def test_eol_lf_with_crlf_in_repo_binary(self):
        wt, basis = self.prepare_tree(_sample_binary, eol="lf-with-crlf-in-repo")
        self.assertContent(wt, basis, _sample_binary, _sample_binary, _sample_binary)

    def test_eol_crlf_with_crlf_in_repo_binary(self):
        wt, basis = self.prepare_tree(_sample_binary, eol="crlf-with-crlf-in-repo")
        self.assertContent(wt, basis, _sample_binary, _sample_binary, _sample_binary)

    # Test text with mixed line endings ("dirty text").
    # This doesn't roundtrip so status always thinks something has changed.

    def test_eol_no_rules_dirty(self):
        wt, basis = self.prepare_tree(_sample_text)
        self.assertContent(
            wt,
            basis,
            _sample_text,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=[],
        )

    def test_eol_exact_dirty(self):
        wt, basis = self.prepare_tree(_sample_text, eol="exact")
        self.assertContent(
            wt,
            basis,
            _sample_text,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=[],
        )

    def test_eol_native_dirty(self):
        wt, basis = self.prepare_tree(_sample_text, eol="native")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=[],
        )

    def test_eol_lf_dirty(self):
        wt, basis = self.prepare_tree(_sample_text, eol="lf")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=[],
        )

    def test_eol_crlf_dirty(self):
        wt, basis = self.prepare_tree(_sample_text, eol="crlf")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=[],
        )

    def test_eol_native_with_crlf_in_repo_dirty(self):
        wt, basis = self.prepare_tree(_sample_text, eol="native-with-crlf-in-repo")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=[],
        )

    def test_eol_lf_with_crlf_in_repo_dirty(self):
        wt, basis = self.prepare_tree(_sample_text, eol="lf-with-crlf-in-repo")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=[],
        )

    def test_eol_crlf_with_crlf_in_repo_dirty(self):
        wt, basis = self.prepare_tree(_sample_text, eol="crlf-with-crlf-in-repo")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=[],
        )

    # Test text with clean line endings, either always lf or always crlf.
    # This selectively roundtrips (based on what's stored in the repo).

    def test_eol_no_rules_clean_lf(self):
        wt, basis = self.prepare_tree(_sample_clean_lf)
        self.assertContent(
            wt,
            basis,
            _sample_clean_lf,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_LF_IN_REPO,
        )

    def test_eol_no_rules_clean_crlf(self):
        wt, basis = self.prepare_tree(_sample_clean_crlf)
        self.assertContent(
            wt,
            basis,
            _sample_clean_crlf,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_CRLF_IN_REPO,
        )

    def test_eol_exact_clean_lf(self):
        wt, basis = self.prepare_tree(_sample_clean_lf, eol="exact")
        self.assertContent(
            wt,
            basis,
            _sample_clean_lf,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_LF_IN_REPO,
        )

    def test_eol_exact_clean_crlf(self):
        wt, basis = self.prepare_tree(_sample_clean_crlf, eol="exact")
        self.assertContent(
            wt,
            basis,
            _sample_clean_crlf,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_CRLF_IN_REPO,
        )

    def test_eol_native_clean_lf(self):
        wt, basis = self.prepare_tree(_sample_clean_lf, eol="native")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_LF_IN_REPO,
        )

    def test_eol_native_clean_crlf(self):
        wt, basis = self.prepare_tree(_sample_clean_crlf, eol="native")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_LF_IN_REPO,
        )

    def test_eol_lf_clean_lf(self):
        wt, basis = self.prepare_tree(_sample_clean_lf, eol="lf")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_LF_IN_REPO,
        )

    def test_eol_lf_clean_crlf(self):
        wt, basis = self.prepare_tree(_sample_clean_crlf, eol="lf")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_LF_IN_REPO,
        )

    def test_eol_crlf_clean_lf(self):
        wt, basis = self.prepare_tree(_sample_clean_lf, eol="crlf")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_LF_IN_REPO,
        )

    def test_eol_crlf_clean_crlf(self):
        wt, basis = self.prepare_tree(_sample_clean_crlf, eol="crlf")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_unix,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_LF_IN_REPO,
        )

    def test_eol_native_with_crlf_in_repo_clean_lf(self):
        wt, basis = self.prepare_tree(_sample_clean_lf, eol="native-with-crlf-in-repo")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_CRLF_IN_REPO,
        )

    def test_eol_native_with_crlf_in_repo_clean_crlf(self):
        wt, basis = self.prepare_tree(
            _sample_clean_crlf, eol="native-with-crlf-in-repo"
        )
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_CRLF_IN_REPO,
        )

    def test_eol_lf_with_crlf_in_repo_clean_lf(self):
        wt, basis = self.prepare_tree(_sample_clean_lf, eol="lf-with-crlf-in-repo")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_CRLF_IN_REPO,
        )

    def test_eol_lf_with_crlf_in_repo_clean_crlf(self):
        wt, basis = self.prepare_tree(_sample_clean_crlf, eol="lf-with-crlf-in-repo")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_CRLF_IN_REPO,
        )

    def test_eol_crlf_with_crlf_in_repo_clean_lf(self):
        wt, basis = self.prepare_tree(_sample_clean_lf, eol="crlf-with-crlf-in-repo")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_CRLF_IN_REPO,
        )

    def test_eol_crlf_with_crlf_in_repo_clean_crlf(self):
        wt, basis = self.prepare_tree(_sample_clean_crlf, eol="crlf-with-crlf-in-repo")
        self.assertContent(
            wt,
            basis,
            _sample_text_on_win,
            _sample_text_on_unix,
            _sample_text_on_win,
            roundtrip_to=_CRLF_IN_REPO,
        )



# Copyright (C) 2010-2012 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License or
# (at your option) a later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Monotone foreign branch support.

Currently only tells the user that Monotone is not supported.
"""

from ... import (
    controldir,
    errors,
    version_info,  # noqa: F401
)


class MonotoneUnsupportedError(errors.UnsupportedVcs):
    vcs = "mtn"

    _fmt = (
        "Monotone branches are not yet supported. "
        "To interoperate with Monotone branches, "
        "use fastimport."
    )


class MonotoneDirFormat(controldir.ControlDirFormat):
    """Monotone directory format."""

    def get_converter(self):
        raise NotImplementedError(self.get_converter)

    def get_format_description(self):
        return "Monotone control directory"

    def initialize_on_transport(self, transport):
        raise errors.UninitializableFormat(self)

    def is_supported(self):
        return False

    def supports_transport(self, transport):
        return False

    def check_support_status(
        self, allow_unsupported, recommend_upgrade=True, basedir=None
    ):
        raise MonotoneUnsupportedError(format=self)

    def open(self, transport):
        # Raise NotBranchError if there is nothing there
        MonotoneProber().probe_transport(transport)
        raise NotImplementedError(self.open)


class MonotoneProber(controldir.Prober):
    @classmethod
    def priority(klass, transport):
        return 100

    @classmethod
    def probe_transport(klass, transport):
        """Our format is present if the transport has a '_MTN/' subdir."""
        if transport.has("_MTN"):
            return MonotoneDirFormat()
        raise errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        return [MonotoneDirFormat()]


controldir.ControlDirFormat.register_prober(MonotoneProber)



# Copyright (C) 2006, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os

from breezy import osutils, tests, workingtree


class TestJoin(tests.TestCaseWithTransport):
    def make_trees(self):
        base_tree = self.make_branch_and_tree("tree", format="development-subtree")
        base_tree.commit("empty commit")
        self.build_tree(["tree/subtree/", "tree/subtree/file1"])
        sub_tree = self.make_branch_and_tree("tree/subtree")
        sub_tree.add("file1", ids=b"file1-id")
        sub_tree.commit("added file1")
        return base_tree, sub_tree

    def check_success(self, path):
        base_tree = workingtree.WorkingTree.open(path)
        self.assertEqual(b"file1-id", base_tree.path2id("subtree/file1"))

    def test_join(self):
        base_tree, sub_tree = self.make_trees()
        self.run_bzr("join tree/subtree")
        self.check_success("tree")

    def test_join_dot(self):
        base_tree, sub_tree = self.make_trees()
        self.run_bzr("join .", working_dir="tree/subtree")
        self.check_success("tree")

    def test_join_error(self):
        base_tree, sub_tree = self.make_trees()
        os.mkdir("tree/subtree2")
        osutils.rename("tree/subtree", "tree/subtree2/subtree")
        self.run_bzr_error(
            ("Cannot join .*subtree.  Parent directory is not versioned",),
            "join tree/subtree2/subtree",
        )
        # disabled because this gives an ugly error at present -- mbp 20070306
        # self.run_bzr_error(
        ##     ('Cannot join .*subtree.  Parent directory is not versioned',),
        # 'join', '--reference', 'tree/subtree2/subtree')
        self.run_bzr_error(("Not a branch:.*subtree2",), "join tree/subtree2")

    def test_join_reference(self):
        """Join can add a reference if --reference is supplied."""
        base_tree, sub_tree = self.make_trees()
        subtree_root_id = sub_tree.path2id("")
        self.run_bzr("join . --reference", working_dir="tree/subtree")
        sub_tree.lock_read()
        self.addCleanup(sub_tree.unlock)
        if sub_tree.supports_setting_file_ids():
            self.assertEqual(b"file1-id", sub_tree.path2id("file1"))
            self.assertEqual("file1", sub_tree.id2path(b"file1-id"))
            self.assertEqual(subtree_root_id, sub_tree.path2id(""))
            self.assertEqual("", sub_tree.id2path(subtree_root_id))
            self.assertEqual(
                sub_tree.path2id("file1"), base_tree.path2id("subtree/file1")
            )

        base_tree.lock_read()
        self.addCleanup(base_tree.unlock)
        self.assertEqual(["subtree"], list(base_tree.iter_references()))
        if base_tree.supports_setting_file_ids():
            self.assertEqual(b"file1-id", sub_tree.path2id("file1"))
            self.assertEqual("file1", sub_tree.id2path(b"file1-id"))
            self.assertEqual(subtree_root_id, base_tree.path2id("subtree"))
            self.assertEqual("subtree", base_tree.id2path(subtree_root_id))

    def test_references_check_repository_support(self):
        """Users are stopped from adding a reference that can't be committed."""
        # in 0.15 the default format has a dirstate workingtree, that can
        # support tree references, but the default repository format
        # cannot.
        self.make_branch_and_tree("tree", format="dirstate")
        self.make_branch_and_tree("tree/subtree")
        out, err = self.run_bzr("join --reference tree/subtree", retcode=3)
        self.assertContainsRe(err, r"Can't join trees")
        self.assertContainsRe(err, r"use brz upgrade")



# Copyright (C) 2005, 2006, 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests of simple versioning operations."""

# TODO: test trying to commit within a directory that is not yet added

import os

from breezy.osutils import pathjoin
from breezy.tests import TestCaseWithTransport
from breezy.trace import mutter
from breezy.workingtree import WorkingTree


class TestMkdir(TestCaseWithTransport):
    def test_mkdir_fails_cleanly(self):
        """'mkdir' fails cleanly when no working tree is available.
        https://bugs.launchpad.net/bzr/+bug/138600.
        """
        # Since there is a safety working tree above us, we create a bare repo
        # here locally.
        self.make_repository(".")
        self.run_bzr(["mkdir", "abc"], retcode=3)
        self.assertPathDoesNotExist("abc")

    def test_mkdir(self):
        """Basic 'brz mkdir' operation."""
        self.make_branch_and_tree(".")
        self.run_bzr(["mkdir", "foo"])
        self.assertTrue(os.path.isdir("foo"))

        self.run_bzr(["mkdir", "foo"], retcode=3)

        wt = WorkingTree.open(".")

        delta = wt.changes_from(wt.basis_tree())

        self.log("delta.added = {!r}".format(delta.added))

        self.assertEqual(len(delta.added), 1)
        self.assertEqual(delta.added[0].path[1], "foo")
        self.assertFalse(delta.modified)

    def test_mkdir_in_subdir(self):
        """'brz mkdir' operation in subdirectory."""
        self.make_branch_and_tree(".")
        self.run_bzr(["mkdir", "dir"])
        self.assertTrue(os.path.isdir("dir"))

        self.log("Run mkdir in subdir")
        self.run_bzr(["mkdir", "subdir"], working_dir="dir")
        self.assertTrue(os.path.isdir("dir/subdir"))

        wt = WorkingTree.open(".")

        delta = wt.changes_from(wt.basis_tree())

        self.log("delta.added = {!r}".format(delta.added))

        self.assertEqual(len(delta.added), 2)
        self.assertEqual(delta.added[0].path[1], "dir")
        self.assertEqual(delta.added[1].path[1], pathjoin("dir", "subdir"))
        self.assertFalse(delta.modified)

    def test_mkdir_w_nested_trees(self):
        """'brz mkdir' with nested trees."""
        self.make_branch_and_tree(".")
        self.make_branch_and_tree("a")
        self.make_branch_and_tree("a/b")

        self.run_bzr(["mkdir", "dir", "a/dir", "a/b/dir"])
        self.assertTrue(os.path.isdir("dir"))
        self.assertTrue(os.path.isdir("a/dir"))
        self.assertTrue(os.path.isdir("a/b/dir"))

        wt = WorkingTree.open(".")
        wt_a = WorkingTree.open("a")
        wt_b = WorkingTree.open("a/b")

        delta = wt.changes_from(wt.basis_tree())
        self.assertEqual(len(delta.added), 1)
        self.assertEqual(delta.added[0].path[1], "dir")
        self.assertFalse(delta.modified)

        delta = wt_a.changes_from(wt_a.basis_tree())
        self.assertEqual(len(delta.added), 1)
        self.assertEqual(delta.added[0].path[1], "dir")
        self.assertFalse(delta.modified)

        delta = wt_b.changes_from(wt_b.basis_tree())
        self.assertEqual(len(delta.added), 1)
        self.assertEqual(delta.added[0].path[1], "dir")
        self.assertFalse(delta.modified)

    def test_mkdir_quiet(self):
        """'brz mkdir --quiet' should not print a status message."""
        self.make_branch_and_tree(".")
        out, err = self.run_bzr(["mkdir", "--quiet", "foo"])
        self.assertEqual("", err)
        self.assertEqual("", out)


class SubdirCommit(TestCaseWithTransport):
    def test_subdir_commit(self):
        """Test committing a subdirectory, and committing a directory."""
        tree = self.make_branch_and_tree(".")
        b = tree.branch
        self.build_tree(["a/", "b/"])

        def set_contents(contents):
            self.build_tree_contents(
                [
                    ("a/one", contents),
                    ("b/two", contents),
                    ("top", contents),
                ]
            )

        set_contents(b"old contents")
        tree.smart_add(["."])
        tree.commit("first revision")
        set_contents(b"new contents")

        mutter("start selective subdir commit")
        self.run_bzr(["commit", "a", "-m", "commit a only"])

        new = b.repository.revision_tree(b.get_rev_id(2))
        new.lock_read()

        def get_text_by_path(tree, path):
            return tree.get_file_text(path)

        self.assertEqual(get_text_by_path(new, "b/two"), b"old contents")
        self.assertEqual(get_text_by_path(new, "top"), b"old contents")
        self.assertEqual(get_text_by_path(new, "a/one"), b"new contents")
        new.unlock()

        # commit from here should do nothing
        self.run_bzr(
            ["commit", ".", "-m", "commit subdir only", "--unchanged"], working_dir="a"
        )
        v3 = b.repository.revision_tree(b.get_rev_id(3))
        v3.lock_read()
        self.assertEqual(get_text_by_path(v3, "b/two"), b"old contents")
        self.assertEqual(get_text_by_path(v3, "top"), b"old contents")
        self.assertEqual(get_text_by_path(v3, "a/one"), b"new contents")
        v3.unlock()

        # commit in subdirectory commits whole tree
        self.run_bzr(["commit", "-m", "commit whole tree from subdir"], working_dir="a")
        v4 = b.repository.revision_tree(b.get_rev_id(4))
        v4.lock_read()
        self.assertEqual(get_text_by_path(v4, "b/two"), b"new contents")
        self.assertEqual(get_text_by_path(v4, "top"), b"new contents")
        v4.unlock()

        # TODO: factor out some kind of assert_tree_state() method



# Copyright (C) 2008-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os
import sys
from io import BytesIO
from textwrap import dedent

from .. import errors, revision, shelf, shelf_ui, tests
from . import features, script


class ExpectShelver(shelf_ui.Shelver):
    """A variant of Shelver that intercepts console activity, for testing."""

    def __init__(
        self,
        work_tree,
        target_tree,
        diff_writer=None,
        auto=False,
        auto_apply=False,
        file_list=None,
        message=None,
        destroy=False,
        reporter=None,
    ):
        shelf_ui.Shelver.__init__(
            self,
            work_tree,
            target_tree,
            diff_writer,
            auto,
            auto_apply,
            file_list,
            message,
            destroy,
            reporter=reporter,
        )
        self.expected = []
        self.diff_writer = BytesIO()

    def expect(self, message, response):
        self.expected.append((message, response))

    def prompt(self, message, choices, default):
        try:
            expected_message, response = self.expected.pop(0)
        except IndexError:
            raise AssertionError("Unexpected prompt: {}".format(message))
        if message != expected_message:
            raise AssertionError("Wrong prompt: {}".format(message))
        if choices != "&yes\n&No\n&finish\n&quit":
            raise AssertionError("Wrong choices: {}".format(choices))
        return response


LINES_AJ = b"a\nb\nc\nd\ne\nf\ng\nh\ni\nj\n"


LINES_ZY = b"z\nb\nc\nd\ne\nf\ng\nh\ni\ny\n"


LINES_AY = b"a\nb\nc\nd\ne\nf\ng\nh\ni\ny\n"


class ShelfTestCase(tests.TestCaseWithTransport):
    def create_shelvable_tree(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/foo", LINES_AJ)])
        tree.add("foo", ids=b"foo-id")
        tree.commit("added foo")
        self.build_tree_contents([("tree/foo", LINES_ZY)])
        return tree


class TestShelver(ShelfTestCase):
    def test_unexpected_prompt_failure(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        e = self.assertRaises(AssertionError, shelver.run)
        self.assertEqual("Unexpected prompt: Shelve?", str(e))

    def test_wrong_prompt_failure(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect("foo", 0)
        e = self.assertRaises(AssertionError, shelver.run)
        self.assertEqual("Wrong prompt: Shelve?", str(e))

    def test_shelve_not_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect("Shelve?", 1)
        shelver.expect("Shelve?", 1)
        # No final shelving prompt because no changes were selected
        shelver.run()
        self.assertFileEqual(LINES_ZY, "tree/foo")

    def test_shelve_diff_no(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect("Shelve?", 0)
        shelver.expect("Shelve?", 0)
        shelver.expect("Shelve 2 change(s)?", 1)
        shelver.run()
        self.assertFileEqual(LINES_ZY, "tree/foo")

    def test_shelve_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect("Shelve?", 0)
        shelver.expect("Shelve?", 0)
        shelver.expect("Shelve 2 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_one_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect("Shelve?", 0)
        shelver.expect("Shelve?", 1)
        shelver.expect("Shelve 1 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AY, "tree/foo")

    def test_shelve_binary_change(self):
        tree = self.create_shelvable_tree()
        self.build_tree_contents([("tree/foo", b"\x00")])
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect("Shelve binary changes?", 0)
        shelver.expect("Shelve 1 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_rename(self):
        tree = self.create_shelvable_tree()
        tree.rename_one("foo", "bar")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve renaming "foo" => "bar"?', 0)
        shelver.expect("Shelve?", 0)
        shelver.expect("Shelve?", 0)
        shelver.expect("Shelve 3 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_deletion(self):
        tree = self.create_shelvable_tree()
        os.unlink("tree/foo")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve removing file "foo"?', 0)
        shelver.expect("Shelve 1 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_creation(self):
        tree = self.make_branch_and_tree("tree")
        tree.commit("add tree root")
        self.build_tree(["tree/foo"])
        tree.add("foo")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve adding file "foo"?', 0)
        shelver.expect("Shelve 1 change(s)?", 0)
        shelver.run()
        self.assertPathDoesNotExist("tree/foo")

    def test_shelve_kind_change(self):
        tree = self.create_shelvable_tree()
        os.unlink("tree/foo")
        os.mkdir("tree/foo")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve changing "foo" from file to directory?', 0)
        shelver.expect("Shelve 1 change(s)?", 0)

    def test_shelve_modify_target(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.create_shelvable_tree()
        os.symlink("bar", "tree/baz")
        tree.add("baz", ids=b"baz-id")
        tree.commit("Add symlink")
        os.unlink("tree/baz")
        os.symlink("vax", "tree/baz")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve changing target of "baz" from "bar" to "vax"?', 0)
        shelver.expect("Shelve 1 change(s)?", 0)
        shelver.run()
        self.assertEqual("bar", os.readlink("tree/baz"))

    def test_shelve_finish(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect("Shelve?", 2)
        shelver.expect("Shelve 2 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_quit(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect("Shelve?", 3)
        self.assertRaises(errors.UserAbort, shelver.run)
        self.assertFileEqual(LINES_ZY, "tree/foo")

    def test_shelve_all(self):
        self.create_shelvable_tree()
        shelver = ExpectShelver.from_args(sys.stdout, all=True, directory="tree")
        try:
            shelver.run()
        finally:
            shelver.finalize()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_filename(self):
        tree = self.create_shelvable_tree()
        self.build_tree(["tree/bar"])
        tree.add("bar")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(), file_list=["bar"])
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve adding file "bar"?', 0)
        shelver.expect("Shelve 1 change(s)?", 0)
        shelver.run()

    def test_shelve_destroy(self):
        tree = self.create_shelvable_tree()
        shelver = shelf_ui.Shelver.from_args(
            sys.stdout, all=True, directory="tree", destroy=True
        )
        self.addCleanup(shelver.finalize)
        shelver.run()
        self.assertIs(None, tree.get_shelf_manager().last_shelf())
        self.assertFileEqual(LINES_AJ, "tree/foo")

    @staticmethod
    def shelve_all(tree, target_revision_id):
        tree.lock_write()
        try:
            target = tree.branch.repository.revision_tree(target_revision_id)
            shelver = shelf_ui.Shelver(tree, target, auto=True, auto_apply=True)
            try:
                shelver.run()
            finally:
                shelver.finalize()
        finally:
            tree.unlock()

    def test_shelve_old_root_preserved(self):
        tree1 = self.make_branch_and_tree("tree1")
        tree1.commit("add root")
        tree1_root_id = tree1.path2id("")
        tree2 = self.make_branch_and_tree("tree2")
        rev2 = tree2.commit("add root")
        self.assertNotEqual(tree1_root_id, tree2.path2id(""))
        tree1.merge_from_branch(tree2.branch, from_revision=revision.NULL_REVISION)
        tree1.commit("merging in tree2")
        self.assertEqual(tree1_root_id, tree1.path2id(""))
        # This is essentially assertNotRaises(InconsistentDelta)
        # With testtools 0.9.9, it can be rewritten as:
        # with ExpectedException(AssertionError,
        #                        'InconsistentDelta not raised'):
        #     with ExpectedException(errors.InconsistentDelta, ''):
        #         self.shelve_all(tree1, rev2)
        e = self.assertRaises(
            AssertionError,
            self.assertRaises,
            errors.InconsistentDelta,
            self.shelve_all,
            tree1,
            rev2,
        )
        self.assertContainsRe("InconsistentDelta not raised", str(e))

    def test_shelve_split(self):
        outer_tree = self.make_branch_and_tree("outer")
        outer_tree.commit("Add root")
        inner_tree = self.make_branch_and_tree("outer/inner")
        rev2 = inner_tree.commit("Add root")
        outer_tree.subsume(inner_tree)
        # This is essentially assertNotRaises(ValueError).
        # The ValueError is 'None is not a valid file id'.
        self.expectFailure(
            "Cannot shelve a join back to the inner tree.",
            self.assertRaises,
            AssertionError,
            self.assertRaises,
            ValueError,
            self.shelve_all,
            outer_tree,
            rev2,
        )


class TestApplyReporter(ShelfTestCase):
    def test_shelve_not_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect("Apply change?", 1)
        shelver.expect("Apply change?", 1)
        # No final shelving prompt because no changes were selected
        shelver.run()
        self.assertFileEqual(LINES_ZY, "tree/foo")

    def test_shelve_diff_no(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect("Apply change?", 0)
        shelver.expect("Apply change?", 0)
        shelver.expect("Apply 2 change(s)?", 1)
        shelver.run()
        self.assertFileEqual(LINES_ZY, "tree/foo")

    def test_shelve_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect("Apply change?", 0)
        shelver.expect("Apply change?", 0)
        shelver.expect("Apply 2 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_binary_change(self):
        tree = self.create_shelvable_tree()
        self.build_tree_contents([("tree/foo", b"\x00")])
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect("Apply binary changes?", 0)
        shelver.expect("Apply 1 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_rename(self):
        tree = self.create_shelvable_tree()
        tree.rename_one("foo", "bar")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect('Rename "bar" => "foo"?', 0)
        shelver.expect("Apply change?", 0)
        shelver.expect("Apply change?", 0)
        shelver.expect("Apply 3 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_deletion(self):
        tree = self.create_shelvable_tree()
        os.unlink("tree/foo")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect('Add file "foo"?', 0)
        shelver.expect("Apply 1 change(s)?", 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, "tree/foo")

    def test_shelve_creation(self):
        tree = self.make_branch_and_tree("tree")
        tree.commit("add tree root")
        self.build_tree(["tree/foo"])
        tree.add("foo")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect('Delete file "foo"?', 0)
        shelver.expect("Apply 1 change(s)?", 0)
        shelver.run()
        self.assertPathDoesNotExist("tree/foo")

    def test_shelve_kind_change(self):
        tree = self.create_shelvable_tree()
        os.unlink("tree/foo")
        os.mkdir("tree/foo")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect('Change "foo" from directory to a file?', 0)
        shelver.expect("Apply 1 change(s)?", 0)

    def test_shelve_modify_target(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.create_shelvable_tree()
        os.symlink("bar", "tree/baz")
        tree.add("baz", ids=b"baz-id")
        tree.commit("Add symlink")
        os.unlink("tree/baz")
        os.symlink("vax", "tree/baz")
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(
            tree, tree.basis_tree(), reporter=shelf_ui.ApplyReporter()
        )
        self.addCleanup(shelver.finalize)
        shelver.expect('Change target of "baz" from "vax" to "bar"?', 0)
        shelver.expect("Apply 1 change(s)?", 0)
        shelver.run()
        self.assertEqual("bar", os.readlink("tree/baz"))


class TestUnshelver(tests.TestCaseWithTransport):
    def create_tree_with_shelf(self):
        tree = self.make_branch_and_tree("tree")
        tree.lock_write()
        try:
            self.build_tree_contents([("tree/foo", LINES_AJ)])
            tree.add("foo", ids=b"foo-id")
            tree.commit("added foo")
            self.build_tree_contents([("tree/foo", LINES_ZY)])
            shelver = shelf_ui.Shelver(
                tree, tree.basis_tree(), auto_apply=True, auto=True
            )
            try:
                shelver.run()
            finally:
                shelver.finalize()
        finally:
            tree.unlock()
        return tree

    def test_unshelve(self):
        tree = self.create_tree_with_shelf()
        tree.lock_write()
        self.addCleanup(tree.unlock)
        manager = tree.get_shelf_manager()
        shelf_ui.Unshelver(tree, manager, 1, True, True, True).run()
        self.assertFileEqual(LINES_ZY, "tree/foo")

    def test_unshelve_args(self):
        tree = self.create_tree_with_shelf()
        unshelver = shelf_ui.Unshelver.from_args(directory="tree")
        try:
            unshelver.run()
        finally:
            unshelver.tree.unlock()
        self.assertFileEqual(LINES_ZY, "tree/foo")
        self.assertIs(None, tree.get_shelf_manager().last_shelf())

    def test_unshelve_args_dry_run(self):
        tree = self.create_tree_with_shelf()
        unshelver = shelf_ui.Unshelver.from_args(directory="tree", action="dry-run")
        try:
            unshelver.run()
        finally:
            unshelver.tree.unlock()
        self.assertFileEqual(LINES_AJ, "tree/foo")
        self.assertEqual(1, tree.get_shelf_manager().last_shelf())

    def test_unshelve_args_preview(self):
        tree = self.create_tree_with_shelf()
        write_diff_to = BytesIO()
        unshelver = shelf_ui.Unshelver.from_args(
            directory="tree", action="preview", write_diff_to=write_diff_to
        )
        try:
            unshelver.run()
        finally:
            unshelver.tree.unlock()
        # The changes were not unshelved.
        self.assertFileEqual(LINES_AJ, "tree/foo")
        self.assertEqual(1, tree.get_shelf_manager().last_shelf())

        # But the diff was written to write_diff_to.
        diff = write_diff_to.getvalue()
        expected = dedent("""\
            @@ -1,4 +1,4 @@
            -a
            +z
             b
             c
             d
            @@ -7,4 +7,4 @@
             g
             h
             i
            -j
            +y

            """)
        self.assertEqualDiff(expected.encode("utf-8"), diff[-len(expected) :])

    def test_unshelve_args_delete_only(self):
        tree = self.make_branch_and_tree("tree")
        manager = tree.get_shelf_manager()
        shelf_file = manager.new_shelf()[1]
        try:
            shelf_file.write(b"garbage")
        finally:
            shelf_file.close()
        unshelver = shelf_ui.Unshelver.from_args(directory="tree", action="delete-only")
        try:
            unshelver.run()
        finally:
            unshelver.tree.unlock()
        self.assertIs(None, manager.last_shelf())

    def test_unshelve_args_invalid_shelf_id(self):
        tree = self.make_branch_and_tree("tree")
        manager = tree.get_shelf_manager()
        shelf_file = manager.new_shelf()[1]
        try:
            shelf_file.write(b"garbage")
        finally:
            shelf_file.close()
        self.assertRaises(
            shelf.InvalidShelfId,
            shelf_ui.Unshelver.from_args,
            directory="tree",
            action="delete-only",
            shelf_id="foo",
        )


class TestUnshelveScripts(TestUnshelver, script.TestCaseWithTransportAndScript):
    def test_unshelve_messages_keep(self):
        self.create_tree_with_shelf()
        self.run_script("""
$ cd tree
$ brz unshelve --keep
2>Using changes with id "1".
2> M  foo
2>All changes applied successfully.
""")

    def test_unshelve_messages_delete(self):
        self.create_tree_with_shelf()
        self.run_script("""
$ cd tree
$ brz unshelve --delete-only
2>Deleted changes with id "1".
""")

    def test_unshelve_messages_apply(self):
        self.create_tree_with_shelf()
        self.run_script("""
$ cd tree
$ brz unshelve --apply
2>Using changes with id "1".
2> M  foo
2>All changes applied successfully.
2>Deleted changes with id "1".
""")

    def test_unshelve_messages_dry_run(self):
        self.create_tree_with_shelf()
        self.run_script("""
$ cd tree
$ brz unshelve --dry-run
2>Using changes with id "1".
2> M  foo
""")



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os
import sys

from breezy import osutils
from breezy.tests import (
    TestCaseWithTransport,
    TestNotApplicable,
    TestSkipped,
    features,
    script,
)
from breezy.workingtree import WorkingTree

_id = b"-id"
a = "a"
b = "b/"
c = "b/c"
d = "d/"
files = (a, b, c, d)


class TestRemove(TestCaseWithTransport):
    def _make_tree_and_add(self, paths):
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            self.build_tree(paths)
            for path in paths:
                file_id = path.replace("/", "_").encode("utf-8") + _id
                tree.add(path, ids=file_id)
        return tree

    def assertFilesDeleted(self, files):
        for f in files:
            f.encode("utf-8") + _id
            self.assertNotInWorkingTree(f)
            self.assertPathDoesNotExist(f)

    def assertFilesUnversioned(self, files):
        for f in files:
            self.assertNotInWorkingTree(f)
            self.assertPathExists(f)

    def changeFile(self, file_name):
        with open(file_name, "ab") as f:
            f.write(b"\nsome other new content!")

    def run_bzr_remove_changed_files(self, files_to_remove, working_dir=None):
        self.run_bzr(["remove"] + list(files_to_remove), working_dir=working_dir)

    def test_remove_new_no_files_specified(self):
        self.make_branch_and_tree(".")
        self.run_bzr_error(["brz: ERROR: No matching files."], "remove --new")
        self.run_bzr_error(["brz: ERROR: No matching files."], "remove --new .")

    def test_remove_no_files_specified(self):
        tree = self._make_tree_and_add(["foo"])
        out, err = self.run_bzr(["rm"])
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertInWorkingTree("foo", tree=tree)
        self.assertPathExists("foo")

    def test_remove_no_files_specified_missing_dir_and_contents(self):
        tree = self._make_tree_and_add(
            ["foo", "dir/", "dir/missing/", "dir/missing/child"]
        )
        self.get_transport(".").delete_tree("dir/missing")
        out, err = self.run_bzr(["rm"])
        self.assertEqual("", out)
        self.assertEqual("removed dir/missing/child\nremoved dir/missing\n", err)
        # non-missing paths not touched:
        self.assertInWorkingTree("foo", tree=tree)
        self.assertPathExists("foo")
        self.assertInWorkingTree("dir", tree=tree)
        self.assertPathExists("dir")
        # missing files unversioned
        self.assertNotInWorkingTree("dir/missing", tree=tree)
        self.assertNotInWorkingTree("dir/missing/child", tree=tree)

    def test_remove_no_files_specified_already_deleted(self):
        tree = self._make_tree_and_add(["foo", "bar"])
        tree.commit("save foo and bar")
        os.unlink("bar")
        self.run_bzr(["rm"])
        self.assertFalse(tree.is_versioned("bar"))
        # Running rm with a deleted file does not error.
        out, err = self.run_bzr(["rm"])
        self.assertEqual("", out)
        self.assertEqual("", err)

    def test_remove_no_files_specified_missing_file(self):
        tree = self._make_tree_and_add(["foo", "bar"])
        os.unlink("bar")
        out, err = self.run_bzr(["rm"])
        self.assertEqual("", out)
        self.assertEqual("removed bar\n", err)
        # non-missing files not touched:
        self.assertInWorkingTree("foo", tree=tree)
        self.assertPathExists("foo")
        # missing files unversioned
        self.assertNotInWorkingTree("bar", tree=tree)

    def test_remove_no_files_specified_missing_link(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self._make_tree_and_add(["foo"])
        os.symlink("foo", "linkname")
        tree.add(["linkname"])
        os.unlink("linkname")
        out, err = self.run_bzr(["rm"])
        self.assertEqual("", out)
        self.assertEqual("removed linkname\n", err)
        # non-missing files not touched:
        self.assertInWorkingTree("foo", tree=tree)
        self.assertPathExists("foo")
        # missing files unversioned
        self.assertNotInWorkingTree("linkname", tree=tree)

    def test_rm_one_file(self):
        self._make_tree_and_add([a])
        self.run_bzr("commit -m 'added a'")
        self.run_bzr("rm a", error_regexes=["deleted a"])
        self.assertFilesDeleted([a])

    def test_remove_one_file(self):
        self._make_tree_and_add([a])
        self.run_bzr("commit -m 'added a'")
        self.run_bzr("remove a", error_regexes=["deleted a"])
        self.assertFilesDeleted([a])

    def test_remove_keep_one_file(self):
        self._make_tree_and_add([a])
        self.run_bzr("remove --keep a", error_regexes=["removed a"])
        self.assertFilesUnversioned([a])

    def test_remove_one_deleted_file(self):
        self._make_tree_and_add([a])
        self.run_bzr("commit -m 'added a'")
        os.unlink(a)
        self.assertInWorkingTree(a)
        self.run_bzr("remove a")
        self.assertNotInWorkingTree(a)

    def test_remove_invalid_files(self):
        self.build_tree(files)
        self.make_branch_and_tree(".")
        self.run_bzr(["remove", ".", "xyz", "abc/def"])

    def test_remove_unversioned_files(self):
        self.build_tree(files)
        self.make_branch_and_tree(".")
        self.run_bzr_remove_changed_files(files)

    def test_remove_changed_files(self):
        self._make_tree_and_add(files)
        self.run_bzr("commit -m 'added files'")
        self.changeFile(a)
        self.changeFile(c)
        self.run_bzr_remove_changed_files(files)

    def test_remove_changed_ignored_files(self):
        self._make_tree_and_add(["a"])
        self.run_bzr(["ignore", "a"])
        self.run_bzr_remove_changed_files(["a"])

    def test_remove_changed_files_from_child_dir(self):
        if sys.platform == "win32":
            raise TestSkipped("Windows unable to remove '.' directory")
        self._make_tree_and_add(files)
        self.run_bzr("commit -m 'added files'")
        self.changeFile(a)
        self.changeFile(c)
        self.run_bzr_remove_changed_files(["../a", "c", ".", "../d"], working_dir="b")
        self.assertNotInWorkingTree(files)
        self.assertPathDoesNotExist(files)

    def test_remove_keep_unversioned_files(self):
        self.build_tree(files)
        self.make_branch_and_tree(".")
        self.run_bzr("remove --keep a", error_regexes=["a is not versioned."])
        self.assertFilesUnversioned(files)

    def test_remove_no_backup_unversioned_files(self):
        self.build_tree(files)
        self.make_branch_and_tree(".")
        script.ScriptRunner().run_script(
            self,
            """
        $ brz remove --no-backup a b/ b/c d/
        2>deleted d
        2>removed b/c (but kept a copy: b/c.~1~)
        2>deleted b
        2>deleted a
        """,
        )
        self.assertFilesDeleted(files)

    def test_remove_deleted_files(self):
        self._make_tree_and_add(files)
        self.run_bzr("commit -m 'added files'")
        my_files = list(files)
        my_files.sort(reverse=True)
        for f in my_files:
            osutils.delete_any(f)
        self.assertInWorkingTree(files)
        self.assertPathDoesNotExist(files)
        self.run_bzr("remove " + " ".join(files))
        self.assertNotInWorkingTree(a)
        self.assertPathDoesNotExist(files)

    def test_remove_non_existing_files(self):
        self._make_tree_and_add([])
        self.run_bzr(["remove", "b"])

    def test_remove_keep_non_existing_files(self):
        self._make_tree_and_add([])
        self.run_bzr("remove --keep b", error_regexes=["b is not versioned."])

    def test_remove_files(self):
        self._make_tree_and_add(files)
        self.run_bzr("commit -m 'added files'")
        self.run_bzr(
            "remove a b b/c d",
            error_regexes=["deleted a", "deleted b", "deleted b/c", "deleted d"],
        )
        self.assertFilesDeleted(files)

    def test_remove_keep_files(self):
        self._make_tree_and_add(files)
        self.run_bzr("commit -m 'added files'")
        self.run_bzr(
            "remove --keep a b b/c d",
            error_regexes=["removed a", "removed b", "removed b/c", "removed d"],
        )
        self.assertFilesUnversioned(files)

    def test_remove_with_new(self):
        self._make_tree_and_add(files)
        self.run_bzr(
            "remove --new --keep",
            error_regexes=["removed a", "removed b", "removed b/c"],
        )
        self.assertFilesUnversioned(files)

    def test_remove_with_new_in_dir1(self):
        tree = self._make_tree_and_add(files)
        self.run_bzr(
            "remove --new --keep b b/c", error_regexes=["removed b", "removed b/c"]
        )
        tree = WorkingTree.open(".")
        self.assertInWorkingTree(a)
        self.assertEqual(tree.path2id(a), a.encode("utf-8") + _id)
        self.assertFilesUnversioned([b, c])

    def test_remove_with_new_in_dir2(self):
        self._make_tree_and_add(files)
        self.run_bzr(
            "remove --new --keep .",
            error_regexes=["removed a", "removed b", "removed b/c"],
        )
        WorkingTree.open(".")
        self.assertFilesUnversioned(files)

    def test_remove_backslash(self):
        # pad.lv/176263
        if os.path.sep == "\\":
            raise TestNotApplicable(
                "unable to add filenames with backslashes where "
                " it is the path separator"
            )
        self.make_branch_and_tree(".")
        self.build_tree(["\\"])
        self.assertEqual("adding \\\n", self.run_bzr("add \\\\")[0])
        self.assertEqual("\\\n", self.run_bzr("ls --versioned")[0])
        self.assertEqual("", self.run_bzr("rm \\\\")[0])
        self.assertEqual("", self.run_bzr("ls --versioned")[0])



# Copyright (C) 2007, 2008, 2009, 2011, 2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import branch as _mod_branch
from breezy import controldir, errors, reconfigure, repository, tests, workingtree
from breezy.bzr import branch as _mod_bzrbranch
from breezy.bzr import vf_repository


class TestReconfigure(tests.TestCaseWithTransport):
    def test_tree_to_branch(self):
        tree = self.make_branch_and_tree("tree")
        reconfiguration = reconfigure.Reconfigure.to_branch(tree.controldir)
        reconfiguration.apply()
        self.assertRaises(errors.NoWorkingTree, workingtree.WorkingTree.open, "tree")

    def test_modified_tree_to_branch(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/file"])
        tree.add("file")
        reconfiguration = reconfigure.Reconfigure.to_branch(tree.controldir)
        self.assertRaises(errors.UncommittedChanges, reconfiguration.apply)
        reconfiguration.apply(force=True)
        self.assertRaises(errors.NoWorkingTree, workingtree.WorkingTree.open, "tree")

    def test_tree_with_pending_merge_to_branch(self):
        tree = self.make_branch_and_tree("tree")
        tree.commit("unchanged")
        other_tree = tree.controldir.sprout("other").open_workingtree()
        other_tree.commit("mergeable commit")
        tree.merge_from_branch(other_tree.branch)
        reconfiguration = reconfigure.Reconfigure.to_branch(tree.controldir)
        self.assertRaises(errors.UncommittedChanges, reconfiguration.apply)
        reconfiguration.apply(force=True)
        self.assertRaises(errors.NoWorkingTree, workingtree.WorkingTree.open, "tree")

    def test_branch_to_branch(self):
        branch = self.make_branch("branch")
        self.assertRaises(
            reconfigure.AlreadyBranch,
            reconfigure.Reconfigure.to_branch,
            branch.controldir,
        )

    def test_repo_to_branch(self):
        repo = self.make_repository("repo")
        reconfiguration = reconfigure.Reconfigure.to_branch(repo.controldir)
        reconfiguration.apply()

    def test_checkout_to_branch(self):
        branch = self.make_branch("branch")
        checkout = branch.create_checkout("checkout")
        reconfiguration = reconfigure.Reconfigure.to_branch(checkout.controldir)
        reconfiguration.apply()
        reconfigured = controldir.ControlDir.open("checkout").open_branch()
        self.assertIs(None, reconfigured.get_bound_location())

    def prepare_lightweight_checkout_to_branch(self):
        branch = self.make_branch("branch")
        checkout = branch.create_checkout("checkout", lightweight=True)
        checkout.commit("first commit", rev_id=b"rev1")
        reconfiguration = reconfigure.Reconfigure.to_branch(checkout.controldir)
        return reconfiguration, checkout

    def test_lightweight_checkout_to_branch(self):
        reconfiguration, checkout = self.prepare_lightweight_checkout_to_branch()
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertEqual(
            checkout_branch.controldir.root_transport.base,
            checkout.controldir.root_transport.base,
        )
        self.assertEqual(b"rev1", checkout_branch.last_revision())
        repo = checkout.controldir.open_repository()
        repo.get_revision(b"rev1")

    def test_lightweight_checkout_to_branch_tags(self):
        reconfiguration, checkout = self.prepare_lightweight_checkout_to_branch()
        checkout.branch.tags.set_tag("foo", b"bar")
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertEqual(b"bar", checkout_branch.tags.lookup_tag("foo"))

    def prepare_lightweight_checkout_to_checkout(self):
        branch = self.make_branch("branch")
        checkout = branch.create_checkout("checkout", lightweight=True)
        reconfiguration = reconfigure.Reconfigure.to_checkout(checkout.controldir)
        return reconfiguration, checkout

    def test_lightweight_checkout_to_checkout(self):
        reconfiguration, checkout = self.prepare_lightweight_checkout_to_checkout()
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertIsNot(checkout_branch.get_bound_location(), None)

    def test_lightweight_checkout_to_checkout_tags(self):
        reconfiguration, checkout = self.prepare_lightweight_checkout_to_checkout()
        checkout.branch.tags.set_tag("foo", b"bar")
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertEqual(b"bar", checkout_branch.tags.lookup_tag("foo"))

    def test_lightweight_conversion_uses_shared_repo(self):
        parent = self.make_branch("parent")
        shared_repo = self.make_repository("repo", shared=True)
        checkout = parent.create_checkout("repo/checkout", lightweight=True)
        reconfigure.Reconfigure.to_tree(checkout.controldir).apply()
        checkout_repo = checkout.controldir.open_branch().repository
        self.assertEqual(
            shared_repo.controldir.root_transport.base,
            checkout_repo.controldir.root_transport.base,
        )

    def test_branch_to_tree(self):
        branch = self.make_branch("branch")
        reconfiguration = reconfigure.Reconfigure.to_tree(branch.controldir)
        reconfiguration.apply()
        branch.controldir.open_workingtree()

    def test_tree_to_tree(self):
        tree = self.make_branch_and_tree("tree")
        self.assertRaises(
            reconfigure.AlreadyTree, reconfigure.Reconfigure.to_tree, tree.controldir
        )

    def test_select_bind_location(self):
        branch = self.make_branch("branch")
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertRaises(
            reconfigure.NoBindLocation, reconfiguration._select_bind_location
        )
        branch.set_parent("http://parent")
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertEqual("http://parent", reconfiguration._select_bind_location())
        branch.set_push_location("sftp://push")
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertEqual("sftp://push", reconfiguration._select_bind_location())
        branch.lock_write()
        try:
            branch.set_bound_location("bzr://foo/old-bound")
            branch.set_bound_location(None)
        finally:
            branch.unlock()
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertEqual("bzr://foo/old-bound", reconfiguration._select_bind_location())
        branch.set_bound_location("bzr://foo/cur-bound")
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertEqual("bzr://foo/cur-bound", reconfiguration._select_bind_location())
        reconfiguration.new_bound_location = "ftp://user-specified"
        self.assertEqual(
            "ftp://user-specified", reconfiguration._select_bind_location()
        )

    def test_select_reference_bind_location(self):
        branch = self.make_branch("branch")
        checkout = branch.create_checkout("checkout", lightweight=True)
        reconfiguration = reconfigure.Reconfigure(checkout.controldir)
        self.assertEqual(branch.base, reconfiguration._select_bind_location())

    def test_tree_to_checkout(self):
        # A tree with no related branches and no supplied bind location cannot
        # become a checkout
        parent = self.make_branch("parent")

        tree = self.make_branch_and_tree("tree")
        reconfiguration = reconfigure.Reconfigure.to_checkout(tree.controldir)
        self.assertRaises(reconfigure.NoBindLocation, reconfiguration.apply)
        # setting a parent allows it to become a checkout
        tree.branch.set_parent(parent.base)
        reconfiguration = reconfigure.Reconfigure.to_checkout(tree.controldir)
        reconfiguration.apply()
        # supplying a location allows it to become a checkout
        tree2 = self.make_branch_and_tree("tree2")
        reconfiguration = reconfigure.Reconfigure.to_checkout(
            tree2.controldir, parent.base
        )
        reconfiguration.apply()

    def test_tree_to_lightweight_checkout(self):
        # A tree with no related branches and no supplied bind location cannot
        # become a checkout
        parent = self.make_branch("parent")

        tree = self.make_branch_and_tree("tree")
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            tree.controldir
        )
        self.assertRaises(reconfigure.NoBindLocation, reconfiguration.apply)
        # setting a parent allows it to become a checkout
        tree.branch.set_parent(parent.base)
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            tree.controldir
        )
        reconfiguration.apply()
        # supplying a location allows it to become a checkout
        tree2 = self.make_branch_and_tree("tree2")
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            tree2.controldir, parent.base
        )
        reconfiguration.apply()

    def test_checkout_to_checkout(self):
        parent = self.make_branch("parent")
        checkout = parent.create_checkout("checkout")
        self.assertRaises(
            reconfigure.AlreadyCheckout,
            reconfigure.Reconfigure.to_checkout,
            checkout.controldir,
        )

    def make_unsynced_checkout(self):
        parent = self.make_branch("parent")
        checkout = parent.create_checkout("checkout")
        checkout.commit("test", rev_id=b"new-commit", local=True)
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            checkout.controldir
        )
        return checkout, parent, reconfiguration

    def test_unsynced_checkout_to_lightweight(self):
        checkout, parent, reconfiguration = self.make_unsynced_checkout()
        self.assertRaises(reconfigure.UnsyncedBranches, reconfiguration.apply)

    def test_synced_checkout_to_lightweight(self):
        checkout, parent, reconfiguration = self.make_unsynced_checkout()
        parent.pull(checkout.branch)
        reconfiguration.apply()
        wt = checkout.controldir.open_workingtree()
        self.assertTrue(parent.repository.has_same_location(wt.branch.repository))
        parent.repository.get_revision(b"new-commit")
        self.assertRaises(
            errors.NoRepositoryPresent, checkout.controldir.open_repository
        )

    def prepare_branch_to_lightweight_checkout(self):
        parent = self.make_branch("parent")
        child = parent.controldir.sprout("child").open_workingtree()
        child.commit("test", rev_id=b"new-commit")
        parent.pull(child.branch)
        child.controldir.destroy_workingtree()
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            child.controldir
        )
        return parent, child, reconfiguration

    def test_branch_to_lightweight_checkout(self):
        parent, child, reconfiguration = self.prepare_branch_to_lightweight_checkout()
        reconfiguration.apply()
        self.assertTrue(reconfiguration._destroy_branch)
        wt = child.controldir.open_workingtree()
        self.assertTrue(parent.repository.has_same_location(wt.branch.repository))
        parent.repository.get_revision(b"new-commit")
        self.assertRaises(errors.NoRepositoryPresent, child.controldir.open_repository)

    def test_branch_to_lightweight_checkout_failure(self):
        parent, child, reconfiguration = self.prepare_branch_to_lightweight_checkout()
        old_Repository_fetch = vf_repository.VersionedFileRepository.fetch
        vf_repository.VersionedFileRepository.fetch = None
        try:
            self.assertRaises(TypeError, reconfiguration.apply)
        finally:
            vf_repository.VersionedFileRepository.fetch = old_Repository_fetch
        child = _mod_branch.Branch.open("child")
        self.assertContainsRe(child.base, "child/$")

    def test_branch_to_lightweight_checkout_fetch_tags(self):
        parent, child, reconfiguration = self.prepare_branch_to_lightweight_checkout()
        child.branch.tags.set_tag("foo", b"bar")
        reconfiguration.apply()
        child = _mod_branch.Branch.open("child")
        self.assertEqual(b"bar", parent.tags.lookup_tag("foo"))

    def test_lightweight_checkout_to_lightweight_checkout(self):
        parent = self.make_branch("parent")
        checkout = parent.create_checkout("checkout", lightweight=True)
        self.assertRaises(
            reconfigure.AlreadyLightweightCheckout,
            reconfigure.Reconfigure.to_lightweight_checkout,
            checkout.controldir,
        )

    def test_repo_to_tree(self):
        repo = self.make_repository("repo")
        reconfiguration = reconfigure.Reconfigure.to_tree(repo.controldir)
        reconfiguration.apply()
        workingtree.WorkingTree.open("repo")

    def test_shared_repo_to_lightweight_checkout(self):
        repo = self.make_repository("repo", shared=True)
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            repo.controldir
        )
        self.assertRaises(reconfigure.NoBindLocation, reconfiguration.apply)
        self.make_branch("branch")
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            repo.controldir, "branch"
        )
        reconfiguration.apply()
        workingtree.WorkingTree.open("repo")
        repository.Repository.open("repo")

    def test_unshared_repo_to_lightweight_checkout(self):
        repo = self.make_repository("repo", shared=False)
        self.make_branch("branch")
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            repo.controldir, "branch"
        )
        reconfiguration.apply()
        workingtree.WorkingTree.open("repo")
        self.assertRaises(
            errors.NoRepositoryPresent, repository.Repository.open, "repo"
        )

    def test_standalone_to_use_shared(self):
        self.build_tree(["root/"])
        tree = self.make_branch_and_tree("root/tree")
        tree.commit("Hello", rev_id=b"hello-id")
        repo = self.make_repository("root", shared=True)
        reconfiguration = reconfigure.Reconfigure.to_use_shared(tree.controldir)
        reconfiguration.apply()
        tree = workingtree.WorkingTree.open("root/tree")
        self.assertTrue(repo.has_same_location(tree.branch.repository))
        self.assertEqual("Hello", repo.get_revision(b"hello-id").message)

    def add_dead_head(self, tree):
        revno, revision_id = tree.branch.last_revision_info()
        tree.commit("Dead head", rev_id=b"dead-head-id")
        tree.branch.set_last_revision_info(revno, revision_id)
        tree.set_last_revision(revision_id)

    def test_standalone_to_use_shared_preserves_dead_heads(self):
        self.build_tree(["root/"])
        tree = self.make_branch_and_tree("root/tree")
        self.add_dead_head(tree)
        tree.commit("Hello", rev_id=b"hello-id")
        repo = self.make_repository("root", shared=True)
        reconfiguration = reconfigure.Reconfigure.to_use_shared(tree.controldir)
        reconfiguration.apply()
        tree = workingtree.WorkingTree.open("root/tree")
        message = repo.get_revision(b"dead-head-id").message
        self.assertEqual("Dead head", message)

    def make_repository_tree(self):
        self.build_tree(["root/"])
        self.make_repository("root", shared=True)
        tree = self.make_branch_and_tree("root/tree")
        reconfigure.Reconfigure.to_use_shared(tree.controldir).apply()
        return workingtree.WorkingTree.open("root/tree")

    def test_use_shared_to_use_shared(self):
        tree = self.make_repository_tree()
        self.assertRaises(
            reconfigure.AlreadyUsingShared,
            reconfigure.Reconfigure.to_use_shared,
            tree.controldir,
        )

    def test_use_shared_to_standalone(self):
        tree = self.make_repository_tree()
        tree.commit("Hello", rev_id=b"hello-id")
        reconfigure.Reconfigure.to_standalone(tree.controldir).apply()
        tree = workingtree.WorkingTree.open("root/tree")
        repo = tree.branch.repository
        self.assertEqual(
            repo.controldir.root_transport.base, tree.controldir.root_transport.base
        )
        self.assertEqual("Hello", repo.get_revision(b"hello-id").message)

    def test_use_shared_to_standalone_preserves_dead_heads(self):
        tree = self.make_repository_tree()
        self.add_dead_head(tree)
        tree.commit("Hello", rev_id=b"hello-id")
        reconfigure.Reconfigure.to_standalone(tree.controldir).apply()
        tree = workingtree.WorkingTree.open("root/tree")
        repo = tree.branch.repository
        self.assertRaises(errors.NoSuchRevision, repo.get_revision, b"dead-head-id")

    def test_standalone_to_standalone(self):
        tree = self.make_branch_and_tree("tree")
        self.assertRaises(
            reconfigure.AlreadyStandalone,
            reconfigure.Reconfigure.to_standalone,
            tree.controldir,
        )

    def make_unsynced_branch_reconfiguration(self):
        parent = self.make_branch_and_tree("parent")
        parent.commit("commit 1")
        child = parent.controldir.sprout("child").open_workingtree()
        child.commit("commit 2")
        return reconfigure.Reconfigure.to_lightweight_checkout(child.controldir)

    def test_unsynced_branch_to_lightweight_checkout_unforced(self):
        reconfiguration = self.make_unsynced_branch_reconfiguration()
        self.assertRaises(reconfigure.UnsyncedBranches, reconfiguration.apply)

    def test_unsynced_branch_to_lightweight_checkout_forced(self):
        reconfiguration = self.make_unsynced_branch_reconfiguration()
        reconfiguration.apply(force=True)

    def make_repository_with_without_trees(self, with_trees):
        repo = self.make_repository("repo", shared=True)
        repo.set_make_working_trees(with_trees)
        return repo

    def test_make_with_trees(self):
        repo = self.make_repository_with_without_trees(False)
        reconfiguration = reconfigure.Reconfigure.set_repository_trees(
            repo.controldir, True
        )
        reconfiguration.apply()
        self.assertIs(True, repo.make_working_trees())

    def test_make_without_trees(self):
        repo = self.make_repository_with_without_trees(True)
        reconfiguration = reconfigure.Reconfigure.set_repository_trees(
            repo.controldir, False
        )
        reconfiguration.apply()
        self.assertIs(False, repo.make_working_trees())

    def test_make_with_trees_already_with_trees(self):
        repo = self.make_repository_with_without_trees(True)
        e = self.assertRaises(
            reconfigure.AlreadyWithTrees,
            reconfigure.Reconfigure.set_repository_trees,
            repo.controldir,
            True,
        )
        self.assertContainsRe(
            str(e), r"Shared repository '.*' already creates working trees."
        )

    def test_make_without_trees_already_no_trees(self):
        repo = self.make_repository_with_without_trees(False)
        e = self.assertRaises(
            reconfigure.AlreadyWithNoTrees,
            reconfigure.Reconfigure.set_repository_trees,
            repo.controldir,
            False,
        )
        self.assertContainsRe(
            str(e), r"Shared repository '.*' already doesn't create working trees."
        )

    def test_repository_tree_reconfiguration_not_supported(self):
        tree = self.make_branch_and_tree("tree")
        e = self.assertRaises(
            reconfigure.ReconfigurationNotSupported,
            reconfigure.Reconfigure.set_repository_trees,
            tree.controldir,
            None,
        )
        self.assertContainsRe(
            str(e), r"Requested reconfiguration of '.*' is not supported."
        )

    def test_lightweight_checkout_to_tree_preserves_reference_locations(self):
        format = controldir.format_registry.make_controldir("1.9")
        format.set_branch_format(_mod_bzrbranch.BzrBranchFormat8())
        tree = self.make_branch_and_tree("tree", format=format)
        tree.branch.set_reference_info(b"file_id", "../location", "path")
        checkout = tree.branch.create_checkout("checkout", lightweight=True)
        reconfiguration = reconfigure.Reconfigure.to_tree(checkout.controldir)
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertEqual(
            ("../location", "path"), checkout_branch.get_reference_info(b"file_id")
        )
